@@ -1,5 +1,5 @@
 --[[
-Copyright (c) 2021, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -226,23 +226,23 @@ local function entry_to_xml(data)
       <disposition>{= data.disposition =}</disposition>
       <dkim>{= data.dkim_disposition =}</dkim>
       <spf>{= data.spf_disposition =}</spf>
-      {% if data.override and data.override ~= '' %}
+      {% if data.override and data.override ~= '' -%}
       <reason><type>{= data.override =}</type></reason>
-      {% endif %}
+      {%- endif %}
     </policy_evaluated>
   </row>
   <identifiers>
     <header_from>{= data.header_from =}</header_from>
   </identifiers>
   <auth_results>
-    {% if data.dkim_results[1] %}
-    {% for d in data.dkim_results[1] %}
+    {% if data.dkim_results[1] -%}
+    {% for d in data.dkim_results -%}
     <dkim>
       <domain>{= d.domain =}</domain>
       <result>{= d.result =}</result>
     </dkim>
-    {% endfor %}
-    {% endif %}
+    {%- endfor %}
+    {%- endif %}
     <spf>
       <domain>{= data.spf_domain =}</domain>
       <result>{= data.spf_result =}</result>
@@ -457,7 +457,7 @@ local function send_reports_by_smtp(opts, reports, finish_cb)
   send_data_in_batches(1)
 end
 
-local function prepare_report(opts, start_time, rep_key)
+local function prepare_report(opts, start_time, end_time, rep_key)
   local rua = get_rua(rep_key)
   local reporting_domain = get_domain(rep_key)
 
@@ -500,7 +500,6 @@ local function prepare_report(opts, start_time, rep_key)
   ret, results = lua_redis.request(redis_params, redis_attrs,
       {'ZRANGE', rep_key, '0', '-1', 'WITHSCORES'})
   local report_entries = {}
-  local end_time = os.time()
   table.insert(report_entries,
       report_header(reporting_domain, start_time, end_time, dmarc_record))
   for i=1,#results,2 do
@@ -568,7 +567,7 @@ local function prepare_report(opts, start_time, rep_key)
   }
 end
 
-local function process_report_date(opts, start_time, date)
+local function process_report_date(opts, start_time, end_time, date)
   local idx_key = redis_prefix(dmarc_settings.reporting.redis_keys.index_prefix, date)
   local ret, results = lua_redis.request(redis_params, redis_attrs,
       {'EXISTS', idx_key})
@@ -599,7 +598,7 @@ local function process_report_date(opts, start_time, date)
 
   local reports = {}
   for _,rep in ipairs(results) do
-    local report = prepare_report(opts, start_time, rep)
+    local report = prepare_report(opts, start_time, end_time, rep)
 
     if report then
       table.insert(reports, report)
@@ -617,7 +616,31 @@ local function process_report_date(opts, start_time, date)
   return reports
 end
 
+
+-- Returns a day before today at 00:00 as unix seconds
+local function yesterday_midnight()
+  local piecewise_time = os.date("!*t")
+  piecewise_time.day = piecewise_time.day - 1 -- Lua allows negative values here
+  piecewise_time.hour = 0
+  piecewise_time.sec = 0
+  piecewise_time.min = 0
+  return os.time(piecewise_time)
+end
+
+-- Returns today time at 00:00 as unix seconds
+local function today_midnight()
+  local piecewise_time = os.date("!*t")
+  piecewise_time.hour = 0
+  piecewise_time.sec = 0
+  piecewise_time.min = 0
+  return os.time(piecewise_time)
+end
+
 local function handler(args)
+  local start_time
+  -- Preserve start time as report sending might take some time
+  local start_collection = today_midnight()
+
   local opts = parser:parse(args)
 
   pool = rspamd_mempool.create()
@@ -653,9 +676,8 @@ local function handler(args)
     'GET', 'rspamd_dmarc_last_collection'
   })
 
-  local start_time
   if not ret or not tonumber(results) then
-    start_time = os.time() - 86400
+    start_time = yesterday_midnight()
   else
     start_time = tonumber(results)
   end
@@ -663,12 +685,8 @@ local function handler(args)
   lua_util.debugm(N, 'previous last report date is %s', start_time)
 
   if not opts.date or #opts.date == 0 then
-    local now = os.time()
     opts.date = {}
-    while now >= start_time do
-      table.insert(opts.date, os.date('!%Y%m%d', now))
-      now = now - 86400
-    end
+    table.insert(opts.date, os.date('!%Y%m%d', yesterday_midnight()))
   end
 
   local ndates = 0
@@ -676,7 +694,7 @@ local function handler(args)
   local all_reports = {}
   for _,date in ipairs(opts.date) do
     lua_util.debugm(N, 'Process date %s', date)
-    local reports_for_date = process_report_date(opts, start_time, date)
+    local reports_for_date = process_report_date(opts, start_time, start_collection, date)
     if #reports_for_date > 0 then
       ndates = ndates + 1
       nreports = nreports + #reports_for_date
@@ -689,7 +707,7 @@ local function handler(args)
 
   local function finish_cb(nsuccess, nfail)
     if not opts.no_opt then
-      lua_util.debugm(N, 'set last report date to %s', os.time())
+      lua_util.debugm(N, 'set last report date to %s', start_collection)
       -- Hack to avoid coroutines + async functions mess: we use async redis call here
       redis_attrs.callback = function()
         logger.messagex('Reporting collection has finished %s dates processed, %s reports: %s completed, %s failed',
@@ -697,7 +715,7 @@ local function handler(args)
       end
       lua_redis.request(redis_params, redis_attrs,
           {'SETEX', 'rspamd_dmarc_last_collection', dmarc_settings.reporting.keys_expire * 2,
-           tostring(os.time())})
+           tostring(start_collection)})
     else
       logger.messagex('Reporting collection has finished %s dates processed, %s reports: %s completed, %s failed',
           ndates, nreports, nsuccess, nfail)
