@@ -1,6 +1,6 @@
 --[[
 Copyright (c) 2016, Andrew Lewis <nerf@judo.za.org>
-Copyright (c) 2016, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -25,6 +25,8 @@ local logger = require "rspamd_logger"
 local util = require "rspamd_util"
 local N = 'milter_headers'
 local lua_util = require "lua_util"
+local lua_maps = require "lua_maps"
+local lua_mime = require "lua_mime"
 local ts = require("tableshape").types
 local E = {}
 
@@ -34,9 +36,10 @@ local settings = {
   remove_upstream_spam_flag = true;
   skip_local = true,
   skip_authenticated = true,
+  skip_all = false,
   local_headers = {},
   authenticated_headers = {},
-  extended_headers_rcpt = {},
+  default_headers_order = nil, -- Insert at the end (set 1 to insert just after the first received)
   routines = {
     ['remove-headers'] = {
       headers = {},
@@ -51,7 +54,8 @@ local settings = {
     ['x-spamd-result'] = {
       header = 'X-Spamd-Result',
       remove = 0,
-      stop_chars = ' '
+      stop_chars = ' ',
+      sort_by = 'score',
     },
     ['x-rspamd-server'] = {
       header = 'X-Rspamd-Server',
@@ -62,6 +66,10 @@ local settings = {
       header = 'X-Rspamd-Queue-Id',
       remove = 0,
     },
+    ['x-rspamd-pre-result'] = {
+      header = 'X-Rspamd-Pre-Result',
+      remove = 0,
+    },
     ['remove-spam-flag'] = {
       header = 'X-Spam',
     },
@@ -102,31 +110,6 @@ local settings = {
     ['authentication-results'] = {
       header = 'Authentication-Results',
       remove = 0,
-      spf_symbols = {
-        pass = 'R_SPF_ALLOW',
-        fail = 'R_SPF_FAIL',
-        softfail = 'R_SPF_SOFTFAIL',
-        neutral = 'R_SPF_NEUTRAL',
-        temperror = 'R_SPF_DNSFAIL',
-        none = 'R_SPF_NA',
-        permerror = 'R_SPF_PERMFAIL',
-      },
-      dkim_symbols = {
-        pass = 'R_DKIM_ALLOW',
-        fail = 'R_DKIM_REJECT',
-        temperror = 'R_DKIM_TEMPFAIL',
-        none = 'R_DKIM_NA',
-        permerror = 'R_DKIM_PERMFAIL',
-      },
-      dmarc_symbols = {
-        pass = 'DMARC_POLICY_ALLOW',
-        permerror = 'DMARC_BAD_POLICY',
-        temperror = 'DMARC_DNSFAIL',
-        none = 'DMARC_NA',
-        reject = 'DMARC_POLICY_REJECT',
-        softfail = 'DMARC_POLICY_SOFTFAIL',
-        quarantine = 'DMARC_POLICY_QUARANTINE',
-      },
       add_smtp_user = true,
       stop_chars = ';',
     },
@@ -145,29 +128,43 @@ local custom_routines = {}
 
 local function milter_headers(task)
 
-  local function skip_wanted(hdr)
+  -- Used to override wanted stuff by means of settings
+  local settings_override = false
 
+  local function skip_wanted(hdr)
+    if settings_override then return true end
+    -- Normal checks
     local function match_extended_headers_rcpt()
       local rcpts = task:get_recipients('smtp')
       if not rcpts then return false end
       local found
       for _, r in ipairs(rcpts) do
         found = false
-        for k, v in pairs(settings.extended_headers_rcpt) do
-          for _, ehr in ipairs(v) do
-            if r[k] == ehr then
-              found = true
-              break
-            end
+        -- Try full addr match
+        if r.addr and r.domain and r.user then
+          if settings.extended_headers_rcpt:get_key(r.addr) then
+            lua_util.debugm(N, task, 'found full addr in recipients for extended headers: %s',
+                r.addr)
+            found = true
+          end
+          -- Try user as plain match
+          if not found and settings.extended_headers_rcpt:get_key(r.user) then
+            lua_util.debugm(N, task, 'found user in recipients for extended headers: %s (%s)',
+                r.user, r.addr)
+            found = true
+          end
+          -- Try @domain to match domain
+          if not found and settings.extended_headers_rcpt:get_key('@' .. r.domain) then
+            lua_util.debugm(N, task, 'found domain in recipients for extended headers: @%s (%s)',
+                r.domain, r.addr)
+            found = true
           end
-          if found then break end
         end
-        if not found then break end
+        if found then break end
       end
       return found
     end
 
-
     if settings.extended_headers_rcpt and match_extended_headers_rcpt() then
       return false
     end
@@ -181,10 +178,19 @@ local function milter_headers(task)
       if task:get_user() ~= nil then return true end
     end
 
+    if settings.skip_all then
+      return true
+    end
+
     return false
 
   end
 
+  -- XXX: fix this crap one day
+  -- routines - are closures that encloses all environment including task
+  -- common - a common environment shared between routines
+  -- add - add headers table (filled by routines)
+  -- remove - remove headers table (filled by routines)
   local routines, common, add, remove = {}, {}, {}, {}
 
   local function add_header(name, value, stop_chars, order)
@@ -219,13 +225,22 @@ local function milter_headers(task)
           }
         end
 
-        table.insert(add[hname],
-            lua_util.fold_header(task, hname, value, stop_chars))
+        if settings.default_headers_order then
+          table.insert(add[hname], {
+            order = settings.default_headers_order,
+            value = lua_util.fold_header(task, hname, value, stop_chars)
+          })
+        else
+          table.insert(add[hname],
+              lua_util.fold_header(task, hname, value, stop_chars))
+        end
+
       end
     end
   end
 
   routines['x-spamd-result'] = function()
+    local local_mod = settings.routines['x-spamd-result']
     if skip_wanted('x-spamd-result') then return end
     if not common.symbols then
       common.symbols = task:get_symbols_all()
@@ -236,21 +251,52 @@ local function milter_headers(task)
     if not common['metric_action'] then
       common['metric_action'] = task:get_metric_action('default')
     end
-    if settings.routines['x-spamd-result'].remove then
-      remove[settings.routines['x-spamd-result'].header] = settings.routines['x-spamd-result'].remove
+    if local_mod.remove then
+      remove[local_mod.header] = local_mod.remove
     end
+
     local buf = {}
-    table.insert(buf, table.concat({
-      'default: ', (common['metric_action'] == 'reject') and 'True' or 'False', ' [',
-      string.format('%.2f', common['metric_score'][1]), ' / ', string.format('%.2f', common['metric_score'][2]), ']'
-    }))
+    local verdict = string.format('default: %s [%.2f / %.2f]',
+        --TODO: (common.metric_action == 'no action') and 'False' or 'True',
+        (common.metric_action == 'reject') and 'True' or 'False',
+        common.metric_score[1], common.metric_score[2])
+    table.insert(buf, verdict)
+
+    -- Deal with symbols
+    table.sort(common.symbols, function(s1, s2)
+      local res
+      if local_mod.sort_by == 'name' then
+        res = s1.name < s2.name
+      else
+        -- inverse order to show important symbols first
+        res = math.abs(s1.score) > math.abs(s2.score)
+      end
+
+      return res
+    end)
+
     for _, s in ipairs(common.symbols) do
-      if not s.options then s.options = {} end
-      table.insert(buf, table.concat({
-        ' ', s.name, '(', string.format('%.2f', s.score), ')[', table.concat(s.options, ','), ']',
-      }))
+      local sym_str = string.format('%s(%.2f)[%s]',
+          s.name, s.score,  table.concat(s.options or {}, ','))
+      table.insert(buf, sym_str)
     end
     add_header('x-spamd-result', table.concat(buf, '; '), ';')
+
+    local has_pr,action,message,module = task:has_pre_result()
+
+    if has_pr then
+      local pr_header = {}
+      if action then
+        table.insert(pr_header, string.format('action=%s', action))
+      end
+      if module then
+        table.insert(pr_header, string.format('module=%s', module))
+      end
+      if message then
+        table.insert(pr_header, message)
+      end
+      add_header('x-rspamd-pre-result', table.concat(pr_header, '; '), ';')
+    end
   end
 
   routines['x-rspamd-queue-id'] = function()
@@ -292,15 +338,17 @@ local function milter_headers(task)
   end
 
   routines['x-rspamd-server'] = function()
+    local local_mod = settings.routines['x-rspamd-server']
     if skip_wanted('x-rspamd-server') then return end
-    if settings.routines['x-rspamd-server'].remove then
-      remove[settings.routines['x-rspamd-server'].header] = settings.routines['x-rspamd-server'].remove
+    if local_mod.remove then
+      remove[local_mod.header] = local_mod.remove
     end
-    local hostname = settings.routines['x-rspamd-server'].hostname
-    add[settings.routines['x-rspamd-server'].header] = hostname and hostname or HOSTNAME
+    local hostname = local_mod.hostname
+    add[local_mod.header] = hostname and hostname or HOSTNAME
   end
 
   routines['x-spamd-bar'] = function()
+    local local_mod = settings.routines['x-spamd-bar']
     if skip_wanted('x-rspamd-bar') then return end
     if not common['metric_score'] then
       common['metric_score'] = task:get_metric_score('default')
@@ -308,21 +356,22 @@ local function milter_headers(task)
     local score = common['metric_score'][1]
     local spambar
     if score <= -1 then
-      spambar = string.rep(settings.routines['x-spamd-bar'].negative, score*-1)
+      spambar = string.rep(local_mod.negative, math.floor(score * -1))
     elseif score >= 1 then
-      spambar = string.rep(settings.routines['x-spamd-bar'].positive, score)
+      spambar = string.rep(local_mod.positive, math.floor(score))
     else
-      spambar = settings.routines['x-spamd-bar'].neutral
+      spambar = local_mod.neutral
     end
-    if settings.routines['x-spamd-bar'].remove then
-      remove[settings.routines['x-spamd-bar'].header] = settings.routines['x-spamd-bar'].remove
+    if local_mod.remove then
+      remove[local_mod.header] = local_mod.remove
     end
     if spambar ~= '' then
-      add[settings.routines['x-spamd-bar'].header] = spambar
+      add[local_mod.header] = spambar
     end
   end
 
   routines['x-spam-level'] = function()
+    local local_mod = settings.routines['x-spam-level']
     if skip_wanted('x-spam-level') then return end
     if not common['metric_score'] then
       common['metric_score'] = task:get_metric_score('default')
@@ -331,10 +380,10 @@ local function milter_headers(task)
     if score < 1 then
       return nil, {}, {}
     end
-    if settings.routines['x-spam-level'].remove then
-      remove[settings.routines['x-spam-level'].header] = settings.routines['x-spam-level'].remove
+    if local_mod.remove then
+      remove[local_mod.header] = local_mod.remove
     end
-    add[settings.routines['x-spam-level'].header] = string.rep(settings.routines['x-spam-level'].char, score)
+    add[local_mod.header] = string.rep(local_mod.char, math.floor(score))
   end
 
   local function spam_header (class, name, value, remove_v)
@@ -363,6 +412,7 @@ local function milter_headers(task)
   end
 
   routines['x-virus'] = function()
+    local local_mod = settings.routines['x-virus']
     if skip_wanted('x-virus') then return end
     if not common.symbols_hash then
       if not common.symbols then
@@ -374,11 +424,11 @@ local function milter_headers(task)
       end
       common.symbols_hash = h
     end
-    if settings.routines['x-virus'].remove then
-      remove[settings.routines['x-virus'].header] = settings.routines['x-virus'].remove
+    if local_mod.remove then
+      remove[local_mod.header] = local_mod.remove
     end
     local virii = {}
-    for _, sym in ipairs(settings.routines['x-virus'].symbols) do
+    for _, sym in ipairs(local_mod.symbols) do
       local s = common.symbols_hash[sym]
       if s then
         if (s.options or E)[1] then
@@ -390,14 +440,14 @@ local function milter_headers(task)
     end
     if #virii > 0 then
       local virusstatus = table.concat(virii, ',')
-      if settings.routines['x-virus'].status_infected then
-        virusstatus = settings.routines['x-virus'].status_infected .. ', ' .. virusstatus
+      if local_mod.status_infected then
+        virusstatus = local_mod.status_infected .. ', ' .. virusstatus
       end
       add_header('x-virus', virusstatus)
     else
       local failed = false
       local fail_reason = 'unknown'
-      for _, sym in ipairs(settings.routines['x-virus'].symbols_fail) do
+      for _, sym in ipairs(local_mod.symbols_fail) do
         local s = common.symbols_hash[sym]
         if s then
           failed = true
@@ -407,13 +457,13 @@ local function milter_headers(task)
         end
       end
       if not failed then
-        if settings.routines['x-virus'].status_clean then
-          add_header('x-virus', settings.routines['x-virus'].status_clean)
+        if local_mod.status_clean then
+          add_header('x-virus', local_mod.status_clean)
         end
       else
-        if settings.routines['x-virus'].status_clean then
+        if local_mod.status_clean then
           add_header('x-virus', string.format('%s(%s)',
-              settings.routines['x-virus'].status_fail, fail_reason))
+              local_mod.status_fail, fail_reason))
         end
       end
     end
@@ -421,6 +471,7 @@ local function milter_headers(task)
 
   routines['x-os-fingerprint'] = function()
     if skip_wanted('x-os-fingerprint') then return end
+    local local_mod = settings.routines['x-os-fingerprint']
 
     local os_string, link_type, uptime_min, distance =
       task:get_mempool():get_variable('os_fingerprint',
@@ -431,9 +482,8 @@ local function milter_headers(task)
     local value = string.format('%s, (up: %i min), (distance %i, link: %s)',
       os_string, uptime_min, distance, link_type)
 
-    if settings.routines['x-os-fingerprint'].remove then
-      remove[settings.routines['x-os-fingerprint'].header]
-        = settings.routines['x-os-fingerprint'].remove
+    if local_mod.remove then
+      remove[local_mod.header] = local_mod.remove
     end
 
     add_header('x-os-fingerprint', value)
@@ -457,6 +507,7 @@ local function milter_headers(task)
       is_spam = 'No'
     end
     spamstatus = is_spam .. ', score=' .. string.format('%.2f', score)
+
     if settings.routines['x-spam-status'].remove then
       remove[settings.routines['x-spam-status'].header] = settings.routines['x-spam-status'].remove
     end
@@ -473,7 +524,8 @@ local function milter_headers(task)
     end
 
     local res = ar.gen_auth_results(task,
-      settings.routines['authentication-results'])
+        lua_util.override_defaults(ar.default_settings,
+            settings.routines['authentication-results']))
 
     if res then
       add_header('authentication-results', res, ';', 1)
@@ -502,7 +554,20 @@ local function milter_headers(task)
     end
   end
 
-  for _, n in ipairs(active_routines) do
+  local routines_enabled = active_routines
+  local user_settings = task:cache_get('settings')
+  if user_settings and user_settings.plugins then
+    user_settings = user_settings.plugins.milter_headers or E
+  end
+
+  if user_settings and type(user_settings.routines) == 'table' then
+    lua_util.debugm(N, task, 'override routines to %s from user settings',
+        user_settings.routines)
+    routines_enabled = user_settings.routines
+    settings_override = true
+  end
+
+  for _, n in ipairs(routines_enabled) do
     local ok, err
     if custom_routines[n] then
       local to_add, to_remove, common_in
@@ -538,9 +603,10 @@ local function milter_headers(task)
   if not next(add) then add = nil end
   if not next(remove) then remove = nil end
   if add or remove then
-    task:set_milter_reply({
-      add_headers = add,
-      remove_headers = remove
+
+    lua_mime.modify_headers(task, {
+      add = add,
+      remove = remove
     })
   end
 end
@@ -553,8 +619,7 @@ local config_schema = ts.shape({
   skip_authenticated = ts.boolean:is_optional(),
   local_headers = ts.array_of(ts.string):is_optional(),
   authenticated_headers = ts.array_of(ts.string):is_optional(),
-  extended_headers_rcpt =
-      (ts.array_of(ts.string) + ts.string / function(s) return {s} end):is_optional(),
+  extended_headers_rcpt = lua_maps.map_schema:is_optional(),
   custom = ts.map_of(ts.string, ts.string):is_optional(),
 }, {
   extra_fields = ts.map_of(ts.string, ts.any)
@@ -583,9 +648,8 @@ local function activate_routine(s)
       have_routine[s] = true
       table.insert(active_routines, s)
       if (opts.routines and opts.routines[s]) then
-        for k, v in pairs(opts.routines[s]) do
-          settings.routines[s][k] = v
-        end
+        settings.routines[s] = lua_util.override_defaults(settings.routines[s],
+            opts.routines[s])
       end
     end
   else
@@ -632,6 +696,10 @@ if type(opts['skip_authenticated']) == 'boolean' then
   settings.skip_authenticated = opts['skip_authenticated']
 end
 
+if type(opts['skip_all']) == 'boolean' then
+  settings.skip_all = opts['skip_all']
+end
+
 for _, s in ipairs(opts['use']) do
   if not have_routine[s] then
     activate_routine(s)
@@ -651,29 +719,8 @@ logger.infox(rspamd_config, 'active routines [%s]',
     table.concat(active_routines, ','))
 
 if opts.extended_headers_rcpt then
-  for _, e in ipairs(opts.extended_headers_rcpt) do
-    if string.find(e, '^[^@]+@[^@]+$') then
-      if not settings.extended_headers_rcpt.addr then
-        settings.extended_headers_rcpt.addr = {}
-      end
-      table.insert(settings.extended_headers_rcpt['addr'], e)
-    elseif string.find(e, '^[^@]+$') then
-      if not settings.extended_headers_rcpt.user then
-        settings.extended_headers_rcpt.user = {}
-      end
-      table.insert(settings.extended_headers_rcpt['user'], e)
-    else
-      local d = string.match(e, '^@([^@]+)$')
-      if d then
-        if not settings.extended_headers_rcpt.domain then
-          settings.extended_headers_rcpt.domain = {}
-        end
-        table.insert(settings.extended_headers_rcpt['domain'], d)
-      else
-        logger.errx(rspamd_config, 'extended_headers_rcpt: unexpected entry: %s', e)
-      end
-    end
-  end
+  settings.extended_headers_rcpt = lua_maps.rspamd_map_add_from_ucl(opts.extended_headers_rcpt,
+      'set', 'Extended headers recipients')
 end
 
 rspamd_config:register_symbol({
